@@ -53,12 +53,14 @@ static const int empty_attribute_list[1] = { None };
 static int api_ver = 0;
 
 /* forward declarations */
-static int driQueryFrameTracking( __DRInativeDisplay * dpy, void * priv,
-    int64_t * sbc, int64_t * missedFrames, float * lastMissedUsage,
-    float * usage );
+static int driQueryFrameTracking( __DRInativeDisplay *dpy, void *priv,
+                                  int64_t *sbc, int64_t *missedFrames,
+                                  float *lastMissedUsage, float *usage );
 
-static void *driCreateNewDrawable(__DRInativeDisplay *dpy, const __GLcontextModes *modes,
-    __DRIid draw, __DRIdrawable *pdraw, int renderType, const int *attrs);
+static void *driCreateNewDrawable(__DRInativeDisplay *dpy,
+                                  const __GLcontextModes *modes,
+                                  __DRIid draw, __DRIdrawable *pdraw,
+                                  int renderType, const int *attrs);
 
 static void driDestroyDrawable(__DRInativeDisplay *dpy, void *drawablePrivate);
 
@@ -321,6 +323,7 @@ static GLboolean DoBindContext(__DRInativeDisplay *dpy,
 
     /* Bind the drawable to the context */
     pcp->driDrawablePriv = pdp;
+    pcp->driReadablePriv = prp;
     pdp->driContextPriv = pcp;
     pdp->refcount++;
     if ( pdp != prp ) {
@@ -337,6 +340,12 @@ static GLboolean DoBindContext(__DRInativeDisplay *dpy,
 	DRM_SPINUNLOCK(&psp->pSAREA->drawable_lock, psp->drawLockID);
     }
 
+    if ((pdp != prp) && (!prp->pStamp || *prp->pStamp != prp->lastStamp)) {
+	DRM_SPINLOCK(&psp->pSAREA->drawable_lock, psp->drawLockID);
+	__driUtilUpdateDrawableInfo(prp);
+	DRM_SPINUNLOCK(&psp->pSAREA->drawable_lock, psp->drawLockID);
+    }
+
     /* Call device-specific MakeCurrent */
     (*psp->DriverAPI.MakeCurrent)(pcp, pdp, prp);
 
@@ -400,14 +409,20 @@ __driUtilUpdateDrawableInfo(__DRIdrawablePrivate *pdp)
     __DRIscreenPrivate *psp;
     __DRIcontextPrivate *pcp = pdp->driContextPriv;
     
-    if (!pcp || (pdp != pcp->driDrawablePriv)) {
-	/* ERROR!!! */
-	return;
+    if (!pcp 
+	|| ((pdp != pcp->driDrawablePriv) && (pdp != pcp->driReadablePriv))) {
+	/* ERROR!!! 
+	 * ...but we must ignore it. There can be many contexts bound to a
+	 * drawable.
+	 */
     }
 
     psp = pdp->driScreenPriv;
     if (!psp) {
 	/* ERROR!!! */
+       _mesa_problem(NULL, "Warning! Possible infinite loop due to bug "
+		     "in file %s, line %d\n",
+		     __FILE__, __LINE__);
 	return;
     }
 
@@ -467,8 +482,27 @@ __driUtilUpdateDrawableInfo(__DRIdrawablePrivate *pdp)
 static void driSwapBuffers( __DRInativeDisplay *dpy, void *drawablePrivate )
 {
     __DRIdrawablePrivate *dPriv = (__DRIdrawablePrivate *) drawablePrivate;
+    drm_clip_rect_t rect;
+
     dPriv->swapBuffers(dPriv);
-    (void) dpy;
+
+    /* Check that we actually have the new damage report method */
+    if (api_ver < 20070105 || dri_interface->reportDamage == NULL)
+	return;
+
+    /* Assume it's affecting the whole drawable for now */
+    rect.x1 = 0;
+    rect.y1 = 0;
+    rect.x2 = rect.x1 + dPriv->w;
+    rect.y2 = rect.y1 + dPriv->h;
+
+    /* Report the damage.  Currently, all our drivers draw directly to the
+     * front buffer, so we report the damage there rather than to the backing
+     * store (if any).
+     */
+    (*dri_interface->reportDamage)(dpy, dPriv->screen, dPriv->draw,
+				   dPriv->x, dPriv->y,
+				   &rect, 1, GL_TRUE);
 }
 
 /**
@@ -545,6 +579,13 @@ static int64_t driSwapBuffersMSC( __DRInativeDisplay * dpy, void *drawablePriv,
                                                            remainder );
 }
 
+static void driCopySubBuffer( __DRInativeDisplay *dpy, void *drawablePrivate,
+			      int x, int y, int w, int h)
+{
+    __DRIdrawablePrivate *dPriv = (__DRIdrawablePrivate *) drawablePrivate;
+    dPriv->driScreenPriv->DriverAPI.CopySubBuffer(dPriv, x, y, w, h);
+    (void) dpy;
+}
 
 /**
  * This is called via __DRIscreenRec's createNewDrawable pointer.
@@ -621,6 +662,9 @@ static void *driCreateNewDrawable(__DRInativeDisplay *dpy,
     pdraw->frameTracking = NULL;
     pdraw->queryFrameTracking = driQueryFrameTracking;
 
+    if (driCompareGLXAPIVersion (20060314) >= 0)
+	pdraw->copySubBuffer = driCopySubBuffer;
+
     /* This special default value is replaced with the configured
      * default value when the drawable is first bound to a direct
      * rendering context. 
@@ -641,8 +685,8 @@ static void *driCreateNewDrawable(__DRInativeDisplay *dpy,
    return (void *) pdp;
 }
 
-static __DRIdrawable *driGetDrawable(__DRInativeDisplay *dpy, __DRIid draw,
-					 void *screenPrivate)
+static __DRIdrawable *
+driGetDrawable(__DRInativeDisplay *dpy, __DRIid draw, void *screenPrivate)
 {
     __DRIscreenPrivate *psp = (__DRIscreenPrivate *) screenPrivate;
 
@@ -653,13 +697,16 @@ static __DRIdrawable *driGetDrawable(__DRInativeDisplay *dpy, __DRIid draw,
     return __driFindDrawable(psp->drawHash, draw);
 }
 
-static void driDestroyDrawable(__DRInativeDisplay *dpy, void *drawablePrivate)
+static void
+driDestroyDrawable(__DRInativeDisplay *dpy, void *drawablePrivate)
 {
     __DRIdrawablePrivate *pdp = (__DRIdrawablePrivate *) drawablePrivate;
-    __DRIscreenPrivate *psp = pdp->driScreenPriv;
-    int scrn = psp->myNum;
+    __DRIscreenPrivate *psp;
+    int scrn;
 
     if (pdp) {
+	psp = pdp->driScreenPriv;
+	scrn = psp->myNum;
         (*psp->DriverAPI.DestroyBuffer)(pdp);
 	if ((*dri_interface->windowExists)(dpy, pdp->draw))
 	    (void)(*dri_interface->destroyDrawable)(dpy, scrn, pdp->draw);
@@ -695,7 +742,8 @@ static void driDestroyDrawable(__DRInativeDisplay *dpy, void *drawablePrivate)
  * This function calls __DriverAPIRec::DestroyContext on \p contextPrivate, calls
  * drmDestroyContext(), and finally frees \p contextPrivate.
  */
-static void driDestroyContext(__DRInativeDisplay *dpy, int scrn, void *contextPrivate)
+static void
+driDestroyContext(__DRInativeDisplay *dpy, int scrn, void *contextPrivate)
 {
     __DRIcontextPrivate  *pcp   = (__DRIcontextPrivate *) contextPrivate;
 
@@ -825,10 +873,14 @@ static void driDestroyScreen(__DRInativeDisplay *dpy, int scrn, void *screenPriv
 	(void)drmUnmap((drmAddress)psp->pSAREA, SAREA_MAX);
 	(void)drmUnmap((drmAddress)psp->pFB, psp->fbSize);
 	_mesa_free(psp->pDevPriv);
-	(void)drmClose(psp->fd);
+	(void)drmCloseOnce(psp->fd);
 	if ( psp->modes != NULL ) {
 	    (*dri_interface->destroyContextModes)( psp->modes );
 	}
+
+	assert(psp->drawHash);
+	drmHashDestroy(psp->drawHash);
+
 	_mesa_free(psp);
     }
 }
