@@ -1,4 +1,3 @@
-/* $XFree86: xc/lib/GL/mesa/src/drv/r300/r300_texstate.c,v 1.3 2003/02/15 22:18:47 dawes Exp $ */
 /*
 Copyright (C) The Weather Channel, Inc.  2002.  All Rights Reserved.
 
@@ -51,14 +50,14 @@ WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 #define VALID_FORMAT(f) ( ((f) <= MESA_FORMAT_RGBA_DXT5			\
 			   || ((f) >= MESA_FORMAT_RGBA_FLOAT32 &&	\
 			       (f) <= MESA_FORMAT_INTENSITY_FLOAT16))	\
-			  && tx_table[f].flag )
+			  && tx_table_le[f].flag )
 
 #define _ASSIGN(entry, format)				\
 	[ MESA_FORMAT_ ## entry ] = { format, 0, 1}
 
 static const struct {
 	GLuint format, filter, flag;
-} tx_table[] = {
+} tx_table_be[] = {
 	/*
 	 * Note that the _REV formats are the same as the non-REV formats.
 	 * This is because the REV and non-REV formats are identical as a
@@ -68,9 +67,9 @@ static const struct {
 	 * byte-swapping), the R300 sees the REV and non-REV formats
 	 * identically.  -- paulus
 	 */
-	    _ASSIGN(RGBA8888, R300_EASY_TX_FORMAT(Y, Z, W, X, W8Z8Y8X8)),
+	    _ASSIGN(RGBA8888, R300_EASY_TX_FORMAT(Z, Y, X, W, W8Z8Y8X8)),
 	    _ASSIGN(RGBA8888_REV, R300_EASY_TX_FORMAT(Y, Z, W, X, W8Z8Y8X8)),
-	    _ASSIGN(ARGB8888, R300_EASY_TX_FORMAT(X, Y, Z, W, W8Z8Y8X8)),
+	    _ASSIGN(ARGB8888, R300_EASY_TX_FORMAT(W, Z, Y, X, W8Z8Y8X8)),
 	    _ASSIGN(ARGB8888_REV, R300_EASY_TX_FORMAT(X, Y, Z, W, W8Z8Y8X8)),
 	    _ASSIGN(RGB888, 0xffffffff),
 	    _ASSIGN(RGB565, R300_EASY_TX_FORMAT(X, Y, Z, ONE, Z5Y6X5)),
@@ -106,6 +105,47 @@ static const struct {
 	    _ASSIGN(INTENSITY_FLOAT16, R300_EASY_TX_FORMAT(X, X, X, X, FL_I16)),
 	    };
 
+static const struct {
+	GLuint format, filter, flag;
+} tx_table_le[] = {
+	    _ASSIGN(RGBA8888, R300_EASY_TX_FORMAT(Y, Z, W, X, W8Z8Y8X8)),
+	    _ASSIGN(RGBA8888_REV, R300_EASY_TX_FORMAT(Z, Y, X, W, W8Z8Y8X8)),
+	    _ASSIGN(ARGB8888, R300_EASY_TX_FORMAT(X, Y, Z, W, W8Z8Y8X8)),
+	    _ASSIGN(ARGB8888_REV, R300_EASY_TX_FORMAT(W, Z, Y, X, W8Z8Y8X8)),
+	    _ASSIGN(RGB888, 0xffffffff),
+	    _ASSIGN(RGB565, R300_EASY_TX_FORMAT(X, Y, Z, ONE, Z5Y6X5)),
+	    _ASSIGN(RGB565_REV, R300_EASY_TX_FORMAT(X, Y, Z, ONE, Z5Y6X5)),
+	    _ASSIGN(ARGB4444, R300_EASY_TX_FORMAT(X, Y, Z, W, W4Z4Y4X4)),
+	    _ASSIGN(ARGB4444_REV, R300_EASY_TX_FORMAT(X, Y, Z, W, W4Z4Y4X4)),
+	    _ASSIGN(ARGB1555, R300_EASY_TX_FORMAT(X, Y, Z, W, W1Z5Y5X5)),
+	    _ASSIGN(ARGB1555_REV, R300_EASY_TX_FORMAT(X, Y, Z, W, W1Z5Y5X5)),
+	    _ASSIGN(AL88, R300_EASY_TX_FORMAT(X, X, X, Y, Y8X8)),
+	    _ASSIGN(AL88_REV, R300_EASY_TX_FORMAT(X, X, X, Y, Y8X8)),
+	    _ASSIGN(RGB332, R300_EASY_TX_FORMAT(X, Y, Z, ONE, Z3Y3X2)),
+	    _ASSIGN(A8, R300_EASY_TX_FORMAT(ZERO, ZERO, ZERO, X, X8)),
+	    _ASSIGN(L8, R300_EASY_TX_FORMAT(X, X, X, ONE, X8)),
+	    _ASSIGN(I8, R300_EASY_TX_FORMAT(X, X, X, X, X8)),
+	    _ASSIGN(CI8, R300_EASY_TX_FORMAT(X, X, X, X, X8)),
+	    _ASSIGN(YCBCR, R300_EASY_TX_FORMAT(X, Y, Z, ONE, G8R8_G8B8)|R300_TX_FORMAT_YUV_MODE ),
+	    _ASSIGN(YCBCR_REV, R300_EASY_TX_FORMAT(X, Y, Z, ONE, G8R8_G8B8)|R300_TX_FORMAT_YUV_MODE),
+	    _ASSIGN(RGB_DXT1, R300_EASY_TX_FORMAT(X, Y, Z, ONE, DXT1)),
+	    _ASSIGN(RGBA_DXT1, R300_EASY_TX_FORMAT(X, Y, Z, W, DXT1)),
+	    _ASSIGN(RGBA_DXT3, R300_EASY_TX_FORMAT(X, Y, Z, W, DXT3)),
+	    _ASSIGN(RGBA_DXT5, R300_EASY_TX_FORMAT(Y, Z, W, X, DXT5)),
+	    _ASSIGN(RGBA_FLOAT32, R300_EASY_TX_FORMAT(Z, Y, X, W, FL_R32G32B32A32)),
+	    _ASSIGN(RGBA_FLOAT16, R300_EASY_TX_FORMAT(Z, Y, X, W, FL_R16G16B16A16)),
+	    _ASSIGN(RGB_FLOAT32, 0xffffffff),
+	    _ASSIGN(RGB_FLOAT16, 0xffffffff),
+	    _ASSIGN(ALPHA_FLOAT32, R300_EASY_TX_FORMAT(ZERO, ZERO, ZERO, X, FL_I32)),
+	    _ASSIGN(ALPHA_FLOAT16, R300_EASY_TX_FORMAT(ZERO, ZERO, ZERO, X, FL_I16)),
+	    _ASSIGN(LUMINANCE_FLOAT32, R300_EASY_TX_FORMAT(X, X, X, ONE, FL_I32)),
+	    _ASSIGN(LUMINANCE_FLOAT16, R300_EASY_TX_FORMAT(X, X, X, ONE, FL_I16)),
+	    _ASSIGN(LUMINANCE_ALPHA_FLOAT32, R300_EASY_TX_FORMAT(X, X, X, Y, FL_I32A32)),
+	    _ASSIGN(LUMINANCE_ALPHA_FLOAT16, R300_EASY_TX_FORMAT(X, X, X, Y, FL_I16A16)),
+	    _ASSIGN(INTENSITY_FLOAT32, R300_EASY_TX_FORMAT(X, X, X, X, FL_I32)),
+	    _ASSIGN(INTENSITY_FLOAT16, R300_EASY_TX_FORMAT(X, X, X, X, FL_I16)),
+	    };
+
 #undef _ASSIGN
 
 
@@ -133,12 +173,18 @@ static void r300SetTexImages(r300ContextPtr rmesa,
 
 	/* Set the hardware texture format
 	 */
-	if (VALID_FORMAT(baseImage->TexFormat->MesaFormat) &&
-	    tx_table[baseImage->TexFormat->MesaFormat].flag) {
-		t->format =
-		    tx_table[baseImage->TexFormat->MesaFormat].format;
-		t->filter |=
-		    tx_table[baseImage->TexFormat->MesaFormat].filter;
+	if (VALID_FORMAT(baseImage->TexFormat->MesaFormat)) {
+		if (_mesa_little_endian()) {
+			t->format =
+			    tx_table_le[baseImage->TexFormat->MesaFormat].format;
+			t->filter |=
+			    tx_table_le[baseImage->TexFormat->MesaFormat].filter;
+		} else {
+			t->format =
+			    tx_table_be[baseImage->TexFormat->MesaFormat].format;
+			t->filter |=
+			    tx_table_be[baseImage->TexFormat->MesaFormat].filter;
+		}
 	} else {
 		_mesa_problem(NULL, "unexpected texture format in %s",
 			      __FUNCTION__);
@@ -172,7 +218,7 @@ static void r300SetTexImages(r300ContextPtr rmesa,
 		if (rmesa->texmicrotile  && (tObj->Target != GL_TEXTURE_RECTANGLE_NV) &&
 		   /* texrect might be able to use micro tiling too in theory? */
 		   (baseImage->Height > 1)) {
-			
+
 			/* allow 32 (bytes) x 1 mip (which will use two times the space
 			   the non-tiled version would use) max if base texture is large enough */
 			if ((numLevels == 1) ||
@@ -182,7 +228,7 @@ static void r300SetTexImages(r300ContextPtr rmesa,
 				t->tile_bits |= R300_TXO_MICRO_TILE;
 			}
 		}
-		
+
 		if (tObj->Target != GL_TEXTURE_RECTANGLE_NV) {
 			/* we can set macro tiling even for small textures, they will be untiled anyway */
 			t->tile_bits |= R300_TXO_MACRO_TILE;
@@ -191,91 +237,85 @@ static void r300SetTexImages(r300ContextPtr rmesa,
 #endif
 
 	for (i = 0; i < numLevels; i++) {
-	  const struct gl_texture_image *texImage;
-	  GLuint size;
-	  
-	  texImage = tObj->Image[0][i + t->base.firstLevel];
-	  if (!texImage)
-	    break;
-	  
-	  /* find image size in bytes */
-	  if (texImage->IsCompressed) {
-	    if ((t->format & R300_TX_FORMAT_DXT1) == R300_TX_FORMAT_DXT1) {
-	      // fprintf(stderr,"DXT 1 %d %08X\n", texImage->Width, t->format);
-	      if ((texImage->Width + 3) < 8) /* width one block */
-		size = texImage->CompressedSize * 4;
-	      else if ((texImage->Width + 3) < 16)
-		size = texImage->CompressedSize * 2;
-	      else size = texImage->CompressedSize;
-	    }
-	    else /* DXT3/5, 16 bytes per block */
-	    {
-	      WARN_ONCE("DXT 3/5 suffers from multitexturing problems!\n");
-	      // fprintf(stderr,"DXT 3/5 %d\n", texImage->Width);
-	      if ((texImage->Width + 3) < 8)
-		size = texImage->CompressedSize * 2;
-	      else size = texImage->CompressedSize;
-	    }
-	    
-	  } else if (tObj->Target == GL_TEXTURE_RECTANGLE_NV) {
-	    size = ((texImage->Width * texelBytes + 63) & ~63) * texImage->Height;
-	    blitWidth = 64 / texelBytes;
-	  } else if (t->tile_bits & R300_TXO_MICRO_TILE) {
-		/* tile pattern is 16 bytes x2. mipmaps stay 32 byte aligned,
-		   though the actual offset may be different (if texture is less than
-		   32 bytes width) to the untiled case */
-		int w = (texImage->Width * texelBytes * 2 + 31) & ~31;
-		size = (w * ((texImage->Height + 1) / 2)) * texImage->Depth;
-		blitWidth = MAX2(texImage->Width, 64 / texelBytes);
-	  } else {
-	    int w = (texImage->Width * texelBytes + 31) & ~31;
-	    size = w * texImage->Height * texImage->Depth;
-	    blitWidth = MAX2(texImage->Width, 64 / texelBytes);
-	  }
-	  assert(size > 0);
-	  
-	  if(0)
-	    fprintf(stderr, "w=%d h=%d d=%d tb=%d intFormat=%d\n", texImage->Width, texImage->Height,
-		    texImage->Depth, texImage->TexFormat->TexelBytes,
-		    texImage->InternalFormat);
-	  
-	  /* Align to 32-byte offset.  It is faster to do this unconditionally
-	   * (no branch penalty).
-	   */
-	  
-	  curOffset = (curOffset + 0x1f) & ~0x1f;
-	  
-	  if (texelBytes) {
-	    t->image[0][i].x = curOffset; /* fix x and y coords up later together with offset */
-	    t->image[0][i].y = 0;
-	    t->image[0][i].width = MIN2(size / texelBytes, blitWidth);
-	    t->image[0][i].height = (size / texelBytes) / t->image[0][i].width;
-	  } else {
-	    t->image[0][i].x = curOffset % R300_BLIT_WIDTH_BYTES;
-	    t->image[0][i].y = curOffset / R300_BLIT_WIDTH_BYTES;
-	    t->image[0][i].width = MIN2(size, R300_BLIT_WIDTH_BYTES);
-	    t->image[0][i].height = size / t->image[0][i].width;
-	  }
-#if 0
-	  /* for debugging only and only  applicable to non-rectangle targets */
-	  assert(size % t->image[0][i].width == 0);
-	  assert(t->image[0][i].x == 0
-		 || (size < R300_BLIT_WIDTH_BYTES
-		     && t->image[0][i].height == 1));
-#endif
-	  
-	  if (0)
-	    fprintf(stderr,
-		    "level %d: %dx%d x=%d y=%d w=%d h=%d size=%d at %d\n",
-		    i, texImage->Width, texImage->Height,
-		    t->image[0][i].x, t->image[0][i].y,
-		    t->image[0][i].width, t->image[0][i].height,
-		    size, curOffset);
-	  
-	  curOffset += size;
-	  
+		const struct gl_texture_image *texImage;
+		GLuint size;
+
+		texImage = tObj->Image[0][i + t->base.firstLevel];
+		if (!texImage)
+			break;
+
+		/* find image size in bytes */
+		if (texImage->IsCompressed) {
+			if ((t->format & R300_TX_FORMAT_DXT1) == R300_TX_FORMAT_DXT1) {
+				// fprintf(stderr,"DXT 1 %d %08X\n", texImage->Width, t->format);
+				if ((texImage->Width + 3) < 8) /* width one block */
+					size = texImage->CompressedSize * 4;
+				else if ((texImage->Width + 3) < 16)
+					size = texImage->CompressedSize * 2;
+				else
+					size = texImage->CompressedSize;
+			} else {
+				/* DXT3/5, 16 bytes per block */
+				WARN_ONCE("DXT 3/5 suffers from multitexturing problems!\n");
+				// fprintf(stderr,"DXT 3/5 %d\n", texImage->Width);
+				if ((texImage->Width + 3) < 8)
+					size = texImage->CompressedSize * 2;
+				else
+					size = texImage->CompressedSize;
+			}
+		} else if (tObj->Target == GL_TEXTURE_RECTANGLE_NV) {
+			size = ((texImage->Width * texelBytes + 63) & ~63) * texImage->Height;
+			blitWidth = 64 / texelBytes;
+		} else if (t->tile_bits & R300_TXO_MICRO_TILE) {
+			/* tile pattern is 16 bytes x2. mipmaps stay 32 byte aligned,
+				though the actual offset may be different (if texture is less than
+				32 bytes width) to the untiled case */
+			int w = (texImage->Width * texelBytes * 2 + 31) & ~31;
+			size = (w * ((texImage->Height + 1) / 2)) * texImage->Depth;
+			blitWidth = MAX2(texImage->Width, 64 / texelBytes);
+		} else {
+			int w = (texImage->Width * texelBytes + 31) & ~31;
+			size = w * texImage->Height * texImage->Depth;
+			blitWidth = MAX2(texImage->Width, 64 / texelBytes);
+		}
+		assert(size > 0);
+
+		if(0)
+			fprintf(stderr, "w=%d h=%d d=%d tb=%d intFormat=%d\n",
+					texImage->Width, texImage->Height,
+					texImage->Depth, texImage->TexFormat->TexelBytes,
+					texImage->InternalFormat);
+
+		/* Align to 32-byte offset.  It is faster to do this unconditionally
+		 * (no branch penalty).
+		 */
+
+		curOffset = (curOffset + 0x1f) & ~0x1f;
+
+		if (texelBytes) {
+			/* fix x and y coords up later together with offset */
+			t->image[0][i].x = curOffset;
+			t->image[0][i].y = 0;
+			t->image[0][i].width = MIN2(size / texelBytes, blitWidth);
+			t->image[0][i].height = (size / texelBytes) / t->image[0][i].width;
+		} else {
+			t->image[0][i].x = curOffset % R300_BLIT_WIDTH_BYTES;
+			t->image[0][i].y = curOffset / R300_BLIT_WIDTH_BYTES;
+			t->image[0][i].width = MIN2(size, R300_BLIT_WIDTH_BYTES);
+			t->image[0][i].height = size / t->image[0][i].width;
+		}
+
+		if (0)
+			fprintf(stderr,
+					"level %d: %dx%d x=%d y=%d w=%d h=%d size=%d at %d\n",
+					i, texImage->Width, texImage->Height,
+					t->image[0][i].x, t->image[0][i].y,
+					t->image[0][i].width, t->image[0][i].height,
+					size, curOffset);
+
+		curOffset += size;
 	}
-	
+
 	/* Align the total size of texture memory block.
 	 */
 	t->base.totalSize =
@@ -315,7 +355,7 @@ static void r300SetTexImages(r300ContextPtr rmesa,
 	} else if (tObj->Target == GL_TEXTURE_CUBE_MAP) {
 		ASSERT(log2Width == log2Height);
 		t->format |= R300_TX_FORMAT_CUBIC_MAP;
-		
+
 		t->format_x |= R200_TEXCOORD_CUBIC_ENV;
 		t->pp_cubic_faces = ((log2Width << R200_FACE_WIDTH_1_SHIFT) |
 				     (log2Height << R200_FACE_HEIGHT_1_SHIFT) |
@@ -331,7 +371,7 @@ static void r300SetTexImages(r300ContextPtr rmesa,
 		ASSERT(log2Width == log2Height);
 		t->format |= R300_TX_FORMAT_CUBIC_MAP;
 	}
-	
+
 	t->size = (((tObj->Image[0][t->base.firstLevel]->Width - 1) << R300_TX_WIDTHMASK_SHIFT)
 			|((tObj->Image[0][t->base.firstLevel]->Height - 1) << R300_TX_HEIGHTMASK_SHIFT))
 			|((numLevels - 1) << R300_TX_MAX_MIP_LEVEL_SHIFT);
