@@ -44,6 +44,9 @@
 #include "vtxfmt.h"
 #include "api_validate.h"
 #include "state.h"
+#include "image.h"
+
+#include "vbo/vbo_context.h"
 
 #define CONV_VB(a, b) rvb->AttribPtr[(a)].size = vb->b->size, \
 			rvb->AttribPtr[(a)].type = GL_FLOAT, \
@@ -71,8 +74,11 @@ void radeon_vb_to_rvb(r300ContextPtr rmesa, struct radeon_vertex_buffer *rvb, st
 	CONV_VB(VERT_ATTRIB_COLOR1, SecondaryColorPtr[0]);
 	CONV_VB(VERT_ATTRIB_FOG, FogCoordPtr);
 	
-	for (i=0; i < MAX_TEXTURE_COORD_UNITS; i++)
+	for (i=0; i < ctx->Const.MaxTextureCoordUnits; i++)
 		CONV_VB(VERT_ATTRIB_TEX0 + i, TexCoordPtr[i]);
+
+	for (i=0; i < MAX_VERTEX_PROGRAM_ATTRIBS; i++)
+		CONV_VB(VERT_ATTRIB_GENERIC0 + i, AttribPtr[VERT_ATTRIB_GENERIC0 + i]);
 	
 	rvb->Primitive = vb->Primitive;
 	rvb->PrimitiveCount = vb->PrimitiveCount;
@@ -84,16 +90,18 @@ void radeon_vb_to_rvb(r300ContextPtr rmesa, struct radeon_vertex_buffer *rvb, st
 
 extern void _tnl_array_init( GLcontext *ctx );
 
-#define CONV(a, b) do { \
-			if (ctx->Array.b.Enabled) { \
-				rmesa->state.VB.AttribPtr[(a)].size = ctx->Array.b.Size; \
-				rmesa->state.VB.AttribPtr[(a)].data = ctx->Array.b.BufferObj->Name ? \
-				(void *)ADD_POINTERS(ctx->Array.b.Ptr, ctx->Array.b.BufferObj->Data) : (void *)ctx->Array.b.Ptr; \
-				rmesa->state.VB.AttribPtr[(a)].stride = ctx->Array.b.StrideB; \
-				rmesa->state.VB.AttribPtr[(a)].type = ctx->Array.b.Type; \
-				enabled |= 1 << (a); \
-			} \
-		   } while (0)
+#define CONV(a, b) \
+    do { \
+	if (ctx->Array.ArrayObj->b.Enabled) { \
+	    rmesa->state.VB.AttribPtr[(a)].size = ctx->Array.ArrayObj->b.Size; \
+	    rmesa->state.VB.AttribPtr[(a)].data = ctx->Array.ArrayObj->b.BufferObj->Name \
+	      ? (void *)ADD_POINTERS(ctx->Array.ArrayObj->b.Ptr, ctx->Array.ArrayObj->b.BufferObj->Data) \
+	      : (void *)ctx->Array.ArrayObj->b.Ptr; \
+	    rmesa->state.VB.AttribPtr[(a)].stride = ctx->Array.ArrayObj->b.StrideB; \
+	    rmesa->state.VB.AttribPtr[(a)].type = ctx->Array.ArrayObj->b.Type; \
+	    enabled |= 1 << (a); \
+	} \
+    } while (0)
 
 static int setup_arrays(r300ContextPtr rmesa, GLint start)
 {
@@ -123,28 +131,18 @@ static int setup_arrays(r300ContextPtr rmesa, GLint start)
 			CONV(i, VertexAttrib[i]);
 	
 	for (i=0; i < VERT_ATTRIB_MAX; i++) {
-		if (enabled & (1 << i)) {
-			rmesa->state.VB.AttribPtr[i].data += rmesa->state.VB.AttribPtr[i].stride * start;
-		} else {
-			def.data = ctx->Current.Attrib[i];
-			memcpy(&rmesa->state.VB.AttribPtr[i], &def, sizeof(struct dt));
-		}
-		
-		/*if(rmesa->state.VB.AttribPtr[i].data == ctx->Current.Attrib[i])
-			fprintf(stderr, "%d is default coord\n", i);*/
+	   rmesa->state.VB.AttribPtr[i].data += rmesa->state.VB.AttribPtr[i].stride * start;
 	}
 	
 	for(i=0; i < VERT_ATTRIB_MAX; i++){
 		if(rmesa->state.VB.AttribPtr[i].type != GL_UNSIGNED_BYTE &&
-			rmesa->state.VB.AttribPtr[i].type != GL_FLOAT){
+#if MESA_LITTLE_ENDIAN
+		   rmesa->state.VB.AttribPtr[i].type != GL_SHORT &&
+#endif
+		   rmesa->state.VB.AttribPtr[i].type != GL_FLOAT){
 			WARN_ONCE("Unsupported format %d at index %d\n", rmesa->state.VB.AttribPtr[i].type, i);
 			return R300_FALLBACK_TCL;
 		}
-		if(rmesa->state.VB.AttribPtr[i].type == GL_UNSIGNED_BYTE &&
-			rmesa->state.VB.AttribPtr[i].size != 4){
-			WARN_ONCE("Unsupported component count for ub colors\n");
-			return R300_FALLBACK_TCL;
-		}
 		
 		/*fprintf(stderr, "%d: ", i);
 		
@@ -173,231 +171,90 @@ static int setup_arrays(r300ContextPtr rmesa, GLint start)
 
 void radeon_init_vtxfmt_a(r300ContextPtr rmesa);
 
-static void radeonDrawElements( GLenum mode, GLsizei count, GLenum type, const GLvoid *c_indices )
-{
-	GET_CURRENT_CONTEXT(ctx);
-	r300ContextPtr rmesa = R300_CONTEXT(ctx);
-	int elt_size;
-	int i;
-	unsigned int min = ~0, max = 0;
-	struct tnl_prim prim;
-	static void *ptr = NULL;
-	static struct r300_dma_region rvb;
-	const GLvoid *indices = c_indices;
-	
-	if (count > 65535) {
-		WARN_ONCE("Too many verts!\n");
-		goto fallback;
-	}
-	
-	if (ctx->Array.ElementArrayBufferObj->Name) {
-		/* use indices in the buffer object */
-		if (!ctx->Array.ElementArrayBufferObj->Data) {
-			_mesa_warning(ctx, "DrawRangeElements with empty vertex elements buffer!");
-			return;
-		}
-		/* actual address is the sum of pointers */
-		indices = (GLvoid *)
-		ADD_POINTERS(ctx->Array.ElementArrayBufferObj->Data, (const GLubyte *) c_indices);
-	}
-	
-	if (!_mesa_validate_DrawElements( ctx, mode, count, type, indices ))
-		return;
-	
-	FLUSH_CURRENT( ctx, 0 );
-	r300ReleaseDmaRegion(rmesa, &rvb, __FUNCTION__);
-	
-	switch (type) {
-	case GL_UNSIGNED_BYTE:
-		for (i=0; i < count; i++) {
-			if(((unsigned char *)indices)[i] < min)
-				min = ((unsigned char *)indices)[i];
-			if(((unsigned char *)indices)[i] > max)
-				max = ((unsigned char *)indices)[i];
-		}
-		
-		elt_size = 2;
-		
-		r300AllocDmaRegion(rmesa, &rvb, count * elt_size, elt_size);
-		rvb.aos_offset = GET_START(&rvb);
-		ptr = rvb.address + rvb.start;
-			
-		for (i=0; i < count; i++)
-			((unsigned short int *)ptr)[i] = ((unsigned char *)indices)[i] - min;
-	break;
-		
-	case GL_UNSIGNED_SHORT:
-		for (i=0; i < count; i++) {
-			if(((unsigned short int *)indices)[i] < min)
-				min = ((unsigned short int *)indices)[i];
-			if(((unsigned short int *)indices)[i] > max)
-				max = ((unsigned short int *)indices)[i];
-		}
-		
-		elt_size = 2;
-		
-		r300AllocDmaRegion(rmesa, &rvb, count * elt_size, elt_size);
-		rvb.aos_offset = GET_START(&rvb);
-		ptr = rvb.address + rvb.start;
-		
-		for (i=0; i < count; i++)
-			((unsigned short int *)ptr)[i] = ((unsigned short int *)indices)[i] - min;
-	break;
-	
-	case GL_UNSIGNED_INT:
-		for (i=0; i < count; i++) {
-			if(((unsigned int *)indices)[i] < min)
-				min = ((unsigned int *)indices)[i];
-			if(((unsigned int *)indices)[i] > max)
-				max = ((unsigned int *)indices)[i];
-		}
-		
-		if (max - min <= 65535)
-			elt_size = 2;
-		else 
-			elt_size = 4;
-		
-		r300AllocDmaRegion(rmesa, &rvb, count * elt_size, elt_size);
-		rvb.aos_offset = GET_START(&rvb);
-		ptr = rvb.address + rvb.start;
-		
-		
-		if (max - min <= 65535)
-			for (i=0; i < count; i++)
-				((unsigned short int *)ptr)[i] = ((unsigned int *)indices)[i] - min;
-		else
-			for (i=0; i < count; i++)
-				((unsigned int *)ptr)[i] = ((unsigned int *)indices)[i] - min;
-	break;
-	
-	default:
-		WARN_ONCE("Unknown elt type!\n");
-	goto fallback;
-	}
-	
-	if (ctx->NewState) 
-		_mesa_update_state( ctx );
-	
-	for (i=_TNL_ATTRIB_MAT_FRONT_AMBIENT; i < _TNL_ATTRIB_INDEX; i++) {
-		rmesa->temp_attrib[i] = TNL_CONTEXT(ctx)->vb.AttribPtr[i];
-		TNL_CONTEXT(ctx)->vb.AttribPtr[i] = &rmesa->dummy_attrib[i];
-	}
-	r300UpdateShaders(rmesa);
-	for (i=_TNL_ATTRIB_MAT_FRONT_AMBIENT; i < _TNL_ATTRIB_INDEX; i++) {
-		TNL_CONTEXT(ctx)->vb.AttribPtr[i] = rmesa->temp_attrib[i];
-	}
-	
-	if (rmesa->state.VB.LockCount) {
-		if (rmesa->state.VB.lock_uptodate == GL_FALSE) {
-			if (setup_arrays(rmesa, rmesa->state.VB.LockFirst))
-				return;
-			
-			rmesa->state.VB.Count = rmesa->state.VB.LockCount;
-			
-			r300ReleaseArrays(ctx);
-			r300EmitArrays(ctx, GL_FALSE);
-			
-			rmesa->state.VB.lock_uptodate = GL_TRUE;
-		}
-		
-		if (min < rmesa->state.VB.LockFirst) {
-			WARN_ONCE("Out of range min %d vs %d!\n", min, rmesa->state.VB.LockFirst);
-			return;
-		}
-		
-		if (max >= rmesa->state.VB.LockFirst + rmesa->state.VB.LockCount) {
-			WARN_ONCE("Out of range max %d vs %d!\n", max, rmesa->state.VB.LockFirst +
-					rmesa->state.VB.LockCount);
-			return;
-		}
-	} else {
-		if (setup_arrays(rmesa, min) >= R300_FALLBACK_TCL)
-			goto fallback;
-		
-		rmesa->state.VB.Count = max - min + 1;
-	}
-	
-	r300UpdateShaderStates(rmesa);
-	
-	rmesa->state.VB.Primitive = &prim;
-	rmesa->state.VB.PrimitiveCount = 1;
-	
-	prim.mode = mode | PRIM_BEGIN | PRIM_END;
-	if (rmesa->state.VB.LockCount)
-		prim.start = min - rmesa->state.VB.LockFirst;
-	else
-		prim.start = 0;
-	prim.count = count;
-	
-	rmesa->state.VB.Elts = ptr;
-	rmesa->state.VB.elt_size = elt_size;
-	
-	r300_run_vb_render(ctx, NULL);
-	
-	if(rvb.buf)
-		radeon_mm_use(rmesa, rvb.buf->id);
-	
-	return;
-	
-	fallback:
-	_tnl_array_init(ctx);
-	_mesa_install_exec_vtxfmt( ctx, &TNL_CONTEXT(ctx)->exec_vtxfmt );
-	CALL_DrawElements(GET_DISPATCH(), (mode, count, type, c_indices));
-	radeon_init_vtxfmt_a(rmesa);
-	_mesa_install_exec_vtxfmt( ctx, &TNL_CONTEXT(ctx)->exec_vtxfmt );
-}
 
-static void radeonDrawRangeElements(GLenum mode, GLuint min, GLuint max, GLsizei count, GLenum type, const GLvoid *c_indices)
+static void radeonDrawRangeElements(GLcontext *ctx,
+				    GLenum mode,
+				    GLuint min,
+				    GLuint max,
+				    GLsizei count, 
+				    GLenum type,
+				    const GLvoid *c_indices)
 {
-	GET_CURRENT_CONTEXT(ctx);
+#if 1
+	return GL_FALSE;
+#else
 	r300ContextPtr rmesa = R300_CONTEXT(ctx);
 	struct tnl_prim prim;
 	int elt_size;
 	int i;
 	void *ptr = NULL;
-	static struct r300_dma_region rvb;
+	struct r300_dma_region rvb;
 	const GLvoid *indices = c_indices;
 	
 	if (count > 65535) {
+		/* TODO */
+		if (mode == GL_POINTS ||
+		    mode == GL_LINES ||
+		    mode == GL_QUADS ||
+		    mode == GL_TRIANGLES) {
+			
+			while (count) {
+				i = r300_get_num_verts(rmesa, MIN2(count, 65535), mode);
+				
+				radeonDrawRangeElements(mode, min, max, i, type, indices);
+				
+				indices += i * _mesa_sizeof_type(type);
+				count -= i;
+			}
+			return GL_TRUE;
+		}
 		WARN_ONCE("Too many verts!\n");
-		goto fallback;
+		return GL_FALSE;
 	}
 	
 	if (ctx->Array.ElementArrayBufferObj->Name) {
 		/* use indices in the buffer object */
 		if (!ctx->Array.ElementArrayBufferObj->Data) {
 			_mesa_warning(ctx, "DrawRangeElements with empty vertex elements buffer!");
-			return;
+			return GL_TRUE;
 		}
 		/* actual address is the sum of pointers */
 		indices = (GLvoid *)
 		ADD_POINTERS(ctx->Array.ElementArrayBufferObj->Data, (const GLubyte *) c_indices);
 	}
 	
-	if (!_mesa_validate_DrawRangeElements( ctx, mode, min, max, count, type, indices ))
-		return;
-	
 	FLUSH_CURRENT( ctx, 0 );
 #ifdef OPTIMIZE_ELTS
 	min = 0;
 #endif
-	r300ReleaseDmaRegion(rmesa, &rvb, __FUNCTION__);
 	
-	switch(type){
+	memset(&rvb, 0, sizeof(rvb));
+	switch (type){
 	case GL_UNSIGNED_BYTE:
+#ifdef FORCE_32BITS_ELTS
+		elt_size = 4;
+#else
 		elt_size = 2;
-		
+#endif	
 		r300AllocDmaRegion(rmesa, &rvb, count * elt_size, elt_size);
 		rvb.aos_offset = GET_START(&rvb);
 		ptr = rvb.address + rvb.start;
 		
+#ifdef FORCE_32BITS_ELTS
+		for(i=0; i < count; i++)
+			((unsigned int *)ptr)[i] = ((unsigned char *)indices)[i] - min;
+#else
 		for(i=0; i < count; i++)
 			((unsigned short int *)ptr)[i] = ((unsigned char *)indices)[i] - min;
+#endif
 	break;
 	
 	case GL_UNSIGNED_SHORT:
+#ifdef FORCE_32BITS_ELTS
+		elt_size = 4;
+#else
 		elt_size = 2;
-		
+#endif	
 #ifdef OPTIMIZE_ELTS
 		if (min == 0 && ctx->Array.ElementArrayBufferObj->Name){
 			ptr = indices;
@@ -408,21 +265,29 @@ static void radeonDrawRangeElements(GLenum mode, GLuint min, GLuint max, GLsizei
 		rvb.aos_offset = GET_START(&rvb);
 		ptr = rvb.address + rvb.start;
 
+#ifdef FORCE_32BITS_ELTS
+		for(i=0; i < count; i++)
+			((unsigned int *)ptr)[i] = ((unsigned short int *)indices)[i] - min;
+#else
 		for(i=0; i < count; i++)
 			((unsigned short int *)ptr)[i] = ((unsigned short int *)indices)[i] - min;
+#endif
 	break;
 	
 	case GL_UNSIGNED_INT:
+#ifdef FORCE_32BITS_ELTS
+		elt_size = 4;
+#else
 		if (max - min <= 65535)
 			elt_size = 2;
 		else 
 			elt_size = 4;
-		
+#endif	
 		r300AllocDmaRegion(rmesa, &rvb, count * elt_size, elt_size);
 		rvb.aos_offset = GET_START(&rvb);
 		ptr = rvb.address + rvb.start;
 		
-		if (max - min <= 65535)
+		if (elt_size == 2)
 			for (i=0; i < count; i++)
 				((unsigned short int *)ptr)[i] = ((unsigned int *)indices)[i] - min;
 		else
@@ -432,7 +297,7 @@ static void radeonDrawRangeElements(GLenum mode, GLuint min, GLuint max, GLsizei
 	
 	default:
 		WARN_ONCE("Unknown elt type!\n");
-	goto fallback;
+		return GL_FALSE;
 	}
 	
 	/* XXX: setup_arrays before state update? */
@@ -440,44 +305,14 @@ static void radeonDrawRangeElements(GLenum mode, GLuint min, GLuint max, GLsizei
 	if (ctx->NewState) 
 		_mesa_update_state( ctx );
 	
-	for (i=_TNL_ATTRIB_MAT_FRONT_AMBIENT; i < _TNL_ATTRIB_INDEX; i++) {
-		rmesa->temp_attrib[i] = TNL_CONTEXT(ctx)->vb.AttribPtr[i];
-		TNL_CONTEXT(ctx)->vb.AttribPtr[i] = &rmesa->dummy_attrib[i];
-	}
 	r300UpdateShaders(rmesa);
-	for (i=_TNL_ATTRIB_MAT_FRONT_AMBIENT; i < _TNL_ATTRIB_INDEX; i++) {
-		TNL_CONTEXT(ctx)->vb.AttribPtr[i] = rmesa->temp_attrib[i];
+
+	if (setup_arrays(rmesa, min) >= R300_FALLBACK_TCL) {
+		r300ReleaseDmaRegion(rmesa, &rvb, __FUNCTION__);
+		return GL_FALSE;
 	}
 
-	if (rmesa->state.VB.LockCount) {
-		if (rmesa->state.VB.lock_uptodate == GL_FALSE) {
-			if (setup_arrays(rmesa, rmesa->state.VB.LockFirst))
-				goto fallback;
-			
-			rmesa->state.VB.Count = rmesa->state.VB.LockCount;
-			
-			r300ReleaseArrays(ctx);
-			r300EmitArrays(ctx, GL_FALSE);
-			
-			rmesa->state.VB.lock_uptodate = GL_TRUE;
-		}
-		
-		if (min < rmesa->state.VB.LockFirst) {
-			WARN_ONCE("Out of range min %d vs %d!\n", min, rmesa->state.VB.LockFirst);
-			goto fallback;
-		}
-		
-		/*if (max >= rmesa->state.VB.LockFirst + rmesa->state.VB.LockCount) {
-			WARN_ONCE("Out of range max %d vs %d!\n", max, rmesa->state.VB.LockFirst +
-					rmesa->state.VB.LockCount);
-			return;
-		}*/
-	} else {
-		if (setup_arrays(rmesa, min) >= R300_FALLBACK_TCL)
-			goto fallback;
-		
-		rmesa->state.VB.Count = max - min + 1;
-	}
+	rmesa->state.VB.Count = max - min + 1;
 	
 	r300UpdateShaderStates(rmesa);
 	
@@ -496,36 +331,36 @@ static void radeonDrawRangeElements(GLenum mode, GLuint min, GLuint max, GLsizei
 	rmesa->state.VB.elt_min = min;
 	rmesa->state.VB.elt_max = max;
 	
-	r300_run_vb_render(ctx, NULL);
+	if (r300_run_vb_render(ctx, NULL)) {
+		r300ReleaseDmaRegion(rmesa, &rvb, __FUNCTION__);
+		return GL_FALSE;
+	}
 	
 	if(rvb.buf)
 		radeon_mm_use(rmesa, rvb.buf->id);
 	
-	return ;
-	
-	fallback:
-	_tnl_array_init(ctx);
-	_mesa_install_exec_vtxfmt( ctx, &TNL_CONTEXT(ctx)->exec_vtxfmt );
-	CALL_DrawRangeElements(GET_DISPATCH(), (mode, min, max, count, type, c_indices));
-	radeon_init_vtxfmt_a(rmesa);
-	_mesa_install_exec_vtxfmt( ctx, &TNL_CONTEXT(ctx)->exec_vtxfmt );
+	r300ReleaseDmaRegion(rmesa, &rvb, __FUNCTION__);
+	return GL_TRUE;
+#endif
 }
 
-static void radeonDrawArrays( GLenum mode, GLint start, GLsizei count )
+static GLboolean radeonDrawArrays( GLcontext *ctx,
+				   GLenum mode, GLint start, GLsizei count )
 {
+#if 1
+	return GL_FALSE;
+#else
 	GET_CURRENT_CONTEXT(ctx);
 	r300ContextPtr rmesa = R300_CONTEXT(ctx);
 	struct tnl_prim prim;
-	int i;
 	
 	if (count > 65535) {
+	   /* TODO: split into multiple draws.
+	    */
 		WARN_ONCE("Too many verts!\n");
-		goto fallback;
+		return GL_FALSE;
 	}
 	
-	if (!_mesa_validate_DrawArrays( ctx, mode, start, count ))
-		return;
-	
 	FLUSH_CURRENT( ctx, 0 );
 	
 	if (ctx->NewState) 
@@ -533,44 +368,12 @@ static void radeonDrawArrays( GLenum mode, GLint start, GLsizei count )
 	
 	/* XXX: setup_arrays before state update? */
 	
-	for (i=_TNL_ATTRIB_MAT_FRONT_AMBIENT; i < _TNL_ATTRIB_INDEX; i++) {
-		rmesa->temp_attrib[i] = TNL_CONTEXT(ctx)->vb.AttribPtr[i];
-		TNL_CONTEXT(ctx)->vb.AttribPtr[i] = &rmesa->dummy_attrib[i];
-	}
 	r300UpdateShaders(rmesa);
-	for (i=_TNL_ATTRIB_MAT_FRONT_AMBIENT; i < _TNL_ATTRIB_INDEX; i++) {
-		TNL_CONTEXT(ctx)->vb.AttribPtr[i] = rmesa->temp_attrib[i];
-	}
 
-	if (rmesa->state.VB.LockCount) {
-		if (rmesa->state.VB.lock_uptodate == GL_FALSE) {
-			if (setup_arrays(rmesa, rmesa->state.VB.LockFirst))
-				return;
-			
-			rmesa->state.VB.Count = rmesa->state.VB.LockCount;
-			
-			r300ReleaseArrays(ctx);
-			r300EmitArrays(ctx, GL_FALSE);
-			
-			rmesa->state.VB.lock_uptodate = GL_TRUE;
-		}
-		
-		if (start < rmesa->state.VB.LockFirst) {
-			WARN_ONCE("Out of range min %d vs %d!\n", start, rmesa->state.VB.LockFirst);
-			goto fallback;
-		}
-		
-		if (start + count - 1 >= rmesa->state.VB.LockFirst + rmesa->state.VB.LockCount) { /* XXX */
-			WARN_ONCE("Out of range max %d vs %d!\n", start + count - 1, rmesa->state.VB.LockFirst +
-					rmesa->state.VB.LockCount);
-			goto fallback;
-		}
-	} else {
-		if (setup_arrays(rmesa, start) >= R300_FALLBACK_TCL)
-			goto fallback;
-		
-		rmesa->state.VB.Count = count;
-	}
+	if (setup_arrays(rmesa, start) >= R300_FALLBACK_TCL)
+		return GL_FALSE;
+
+	rmesa->state.VB.Count = count;
 
 	r300UpdateShaderStates(rmesa);
 	
@@ -589,70 +392,75 @@ static void radeonDrawArrays( GLenum mode, GLint start, GLsizei count )
 	rmesa->state.VB.elt_min = 0;
 	rmesa->state.VB.elt_max = 0;
 	
-	r300_run_vb_render(ctx, NULL);
+	if (r300_run_vb_render(ctx, NULL))
+	   return GL_FALSE;
 
-	return ;
-	
-	fallback:
-	_tnl_array_init(ctx);
-	_mesa_install_exec_vtxfmt( ctx, &TNL_CONTEXT(ctx)->exec_vtxfmt );
-	CALL_DrawArrays(GET_DISPATCH(), (mode, start, count));
-	radeon_init_vtxfmt_a(rmesa);
-	_mesa_install_exec_vtxfmt( ctx, &TNL_CONTEXT(ctx)->exec_vtxfmt );
+	return GL_TRUE;
+#endif
 }
 
+static void radeon_draw_prims( GLcontext *ctx,
+			       const struct gl_client_array *arrays[],
+			       const struct _mesa_prim *prim,
+			       GLuint nr_prims,
+			       const struct _mesa_index_buffer *ib,
+			       GLuint min_index,
+			       GLuint max_index)
+{
+	if (ib == NULL) {
+		for (i = 0; i < nr_prims; i++) {
+			if (!radeonDrawArrays(ctx,
+					      prim->mode,
+					      prim->start,
+					      prim->count)) {
+				/* Fallback
+				 */
+				_tnl_draw_prims(ctx,
+						arrays,
+						prim + i,
+						nr_prims - i,
+						ib, 
+						min_index,
+						max_index);
+				return;
+			}
+		}
+	} else {
+		for (i = 0; i < nr_prims; i++) {
+			if (!radeonDrawRangeElements(ctx,
+						     prim->mode,
+						     min_index,
+						     max_index,
+						     prim->count,
+						     ib->types,
+						     ib->ptr)) {
+				/* Fallback
+				 */
+				_tnl_draw_prims(ctx,
+						arrays,
+						prim + i,
+						nr_prims - i,
+						ib, 
+						min_index,
+						max_index);
+				return;
+			}
+		}
+	}
+}
+	   
 void radeon_init_vtxfmt_a(r300ContextPtr rmesa)
 {
 	GLcontext *ctx;
-	GLvertexformat *vfmt;
-	
-	ctx = rmesa->radeon.glCtx; 
-	vfmt = (GLvertexformat *)ctx->TnlModule.Current;
-   
-	vfmt->DrawElements = radeonDrawElements;
-	vfmt->DrawArrays = radeonDrawArrays;
-	vfmt->DrawRangeElements = radeonDrawRangeElements;
+	struct vbo_context *vbo = vbo_context(ctx);
 	
+	vbo->draw_prims = radeon_draw_prims;
 }
+
 #endif
 
 #ifdef HW_VBOS
 
-#if 0
-static void radeonLockArraysEXT(GLcontext *ctx, GLint first, GLsizei count)
-{
-	r300ContextPtr rmesa = R300_CONTEXT(ctx);
-	
-	/* Only when CB_DPATH is defined.
-	   r300Clear tampers over the aos setup without it.
-	   (r300ResetHwState cannot call r300EmitArrays)
-	 */
-#ifndef CB_DPATH
-	first = 0; count = 0;
-#endif
-	
-	if (first < 0 || count <= 0) {
-		rmesa->state.VB.LockFirst = 0;
-		rmesa->state.VB.LockCount = 0;
-		rmesa->state.VB.lock_uptodate = GL_FALSE;
-		return ;
-	}
-	
-	rmesa->state.VB.LockFirst = first;
-	rmesa->state.VB.LockCount = count;
-	rmesa->state.VB.lock_uptodate = GL_FALSE;
-}
-
-static void radeonUnlockArraysEXT(GLcontext *ctx)
-{
-	r300ContextPtr rmesa = R300_CONTEXT(ctx);
-	
-	rmesa->state.VB.LockFirst = 0;
-	rmesa->state.VB.LockCount = 0;
-	rmesa->state.VB.lock_uptodate = GL_FALSE;
-}
-#endif
-
 static struct gl_buffer_object *
 r300NewBufferObject(GLcontext *ctx, GLuint name, GLenum target )
 {
@@ -809,15 +617,14 @@ void r300_evict_vbos(GLcontext *ctx, int amount)
 	r300ContextPtr rmesa = R300_CONTEXT(ctx);
 	struct _mesa_HashTable *hash = ctx->Shared->BufferObjects;
 	GLuint k = _mesa_HashFirstEntry(hash);
-	struct gl_buffer_object *obj;
-	struct r300_buffer_object *r300_obj;
-	GLvoid *data;
 	
 	while (amount > 0 && k) {
-		obj = (struct gl_buffer_object *) _mesa_HashLookup(hash, k);
-		r300_obj = (struct r300_buffer_object *) obj;
+		struct gl_buffer_object *obj = _mesa_lookup_bufferobj(ctx, k);
+		struct r300_buffer_object *r300_obj
+			= (struct r300_buffer_object *) obj;
 		
 		if (obj->OnCard && obj->Size) {
+			GLvoid *data;
 			obj->Data = _mesa_malloc(obj->Size);
 			
 			data = radeon_mm_map(rmesa, r300_obj->id, RADEON_MM_R);
@@ -844,9 +651,6 @@ void r300_init_vbo_funcs(struct dd_function_table *functions)
 	functions->MapBuffer = r300MapBuffer;
 	functions->UnmapBuffer = r300UnmapBuffer;
 	functions->DeleteBuffer = r300DeleteBuffer;
-	
-	/*functions->LockArraysEXT = radeonLockArraysEXT;
-	functions->UnlockArraysEXT = radeonUnlockArraysEXT;*/
 }
 
 #endif
