@@ -928,6 +928,9 @@ _mesa_generate_mipmap(GLcontext *ctx, GLenum target,
          return;
       }
 
+      if (dstImage->ImageOffsets)
+         _mesa_free(dstImage->ImageOffsets);
+
       /* Free old image data */
       if (dstImage->Data)
          ctx->Driver.FreeTexImageData(ctx, dstImage);
